/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.ws.notification;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.objectweb.petals.ws.addressing.EndpointReference;
import org.objectweb.petals.ws.fault.WsnFault;
import org.objectweb.petals.ws.notification.handlers.request.WSNHandler;
import org.objectweb.petals.ws.topic.Topic;
import org.objectweb.petals.ws.topic.TopicNamespace;
import org.objectweb.petals.ws.topic.TopicSet;
import org.objectweb.petals.ws.topic.WstConstants;

/**
 * @author chamerling - eBM WebSourcing
 *
 */
public class WsnManager {

    /**
     * It's me !
     */
    protected EndpointReference producerEPR;

    protected Set<QName> allTopicNames;

    protected String subscriptionPath = null;

    /**
     * Map of EPR to current subscriptions
     */
    protected Map<EndpointReference, Subscription> subscriptionsByEPR;;

    /**
     * Hierarchical topic
     */
    protected TopicSet topicSet;

    /**
     * Map of handlers
     */
    protected Map<String, WSNHandler> wsnHandlers;

    /**
     * Creates a new instance of {@link WsnManager}
     *
     */
    public WsnManager() {
        super();
        this.allTopicNames = new HashSet<QName>();
        this.subscriptionsByEPR = new HashMap<EndpointReference, Subscription>();
        this.topicSet = new TopicSet();
        this.wsnHandlers = new HashMap<String, WSNHandler>();
    }

    /**
     *
     * @param wsnHandler
     */
    public void addHandler(WSNHandler wsnHandler) {
        wsnHandler.setManager(this);
        this.wsnHandlers.put(wsnHandler.getActionURI(), wsnHandler);
    }

    /**
     *
     * @param actionURI
     * @return
     */
    public WSNHandler getHandler(String actionURI) {
        return this.wsnHandlers.get(actionURI);
    }

    /**
     * Add a subscription
     *
     * @param subscription
     */
    public synchronized void addSubscription(Subscription subscription) {
        EndpointReference epr = subscription.getConsumerEPR();
        this.subscriptionsByEPR.put(epr, subscription);
    }

    /**
     * Add a new topic.
     *
     * @param topicName
     * @return
     * @throws WsnFault
     */
    public Topic addTopic(QName topicName) throws WsnFault {
        String namespace = topicName.getNamespaceURI();
        TopicNamespace topicSpace = getTopicNamespace(namespace);

        if (topicSpace == null) {
            topicSpace = addTopicNamespace(namespace);
        }

        String localName = topicName.getLocalPart();
        Topic topic = new Topic(localName, topicSpace);
        topicSpace.addTopic(topic);
        allTopicNames.add(topicName);

        return topic;
    }

    /**
     *
     * @param namespace
     * @return
     * @throws WsnFault
     */
    public TopicNamespace addTopicNamespace(String namespace) throws WsnFault {
        TopicNamespace tns = new TopicNamespace(namespace);
        tns.setName("ResourcePropertiesTopicSpace");
        topicSet.addTopicNamespace(tns);
        return tns;
    }

    /**
     * Get the current notification message for the given topic
     *
     * @param topicPath
     * @return the current message of the topic
     */
    public NotificationMessage getCurrentMessage(QName topicPath)
            throws WsnFault {
        Topic topic = getTopic(topicPath);

        if (topic == null) {
            throw new WsnFault("Topic not found");
        }

        NotificationMessage message = topic.getCurrentMessage();

        if (message == null) {
            throw new WsnFault("No message in topic");
        }
        return message;
    }

    /**
     * Get the topic
     *
     * @param topicName
     * @return
     */
    public Topic getTopic(QName topicName) {
        TopicNamespace topicSpace = getTopicNamespace(topicName
                .getNamespaceURI());

        if (topicSpace == null)
            return null;

        return topicSpace.getTopic(topicName.getLocalPart());
    }

    public QName[] getTopicExpression() {
        QName[] array = new QName[allTopicNames.size()];
        return (QName[]) allTopicNames.toArray(array);
    }

    public String[] getTopicExpressionDialect() {
        return new String[] { WstConstants.CONCRETE_TOPIC_URI };
    }

    public TopicNamespace getTopicNamespace(String namespace) {
        return getTopicSet().getTopicNamespace(namespace);
    }

    public TopicSet getTopicSet() {
        return topicSet;
    }

    /**
     * There is already a subcriber for the given EPR ?
     *
     * @param subscriptionEPR
     * @return
     */
    protected boolean hasSubscription(EndpointReference subscriptionEPR) {
        return this.subscriptionsByEPR.containsKey(subscriptionEPR);
    }

    /**
     * There is already a topic with the given name ?
     *
     * @param topicName
     * @return
     */
    public boolean hasTopic(QName topicName) {
        return allTopicNames.contains(topicName);
    }

    /**
     * Publish
     *
     * @param topicName
     * @param content
     * @throws WsnFault
     */
    public void publish(QName topicName, OMElement content) throws WsnFault {
        publish(topicName, new OMElement[] { content });
    }

    /**
     *
     * @param topicName
     * @param content
     * @throws WsnFault
     */
    public void publish(QName topicName, OMElement[] content) throws WsnFault {
        NotificationMessage message = new NotificationMessage();
        for (OMElement element : content) {
            message.addMessageContent(element);
        }
        message.setTopic(topicName);

        //
        // give the message to each subscription so it can decide if it
        // should be sent or not
        //
        Iterator<Subscription> i = getSubscriptions().iterator();
        while (i.hasNext()) {
            Subscription sub = i.next();
            sub.publish(message);
        }

        //
        // if a topic was used, record the message as the 'current' message
        // for the topic (will be returned by getCurrentMessage())
        //
        if (topicName != null) {
            Topic topic = getTopic(topicName);
            if (topic != null) {
                topic.setCurrentMessage(message);
            }
        }
    }

    /**
     *
     * @return
     */
    protected synchronized Collection<Subscription> getSubscriptions() {
        return Collections.unmodifiableCollection(subscriptionsByEPR.values());
    }

    /**
     *
     * @param epr
     */
    public synchronized void removeSubscription(EndpointReference epr) {
        subscriptionsByEPR.remove(epr);
    }

    /**
     * @return the producerEPR
     */
    public EndpointReference getProducerEPR() {
        return this.producerEPR;
    }

    /**
     * @param producerEPR
     *            the producerEPR to set
     */
    public void setProducerEPR(EndpointReference producerEPR) {
        this.producerEPR = producerEPR;
    }
}
